-- Engineer: Mario A.
-- Fixed-Point LMS Algorithm
-- Data In Q23, Coefficients Q15, Data Out Q23

library IEEE;
use IEEE.STD_LOGIC_1164.ALL; 
use ieee.numeric_std.all;

entity LMS is
    generic(
              NBTAP   : natural := 32;  
              DSIZE   : natural := 24;
              COESIZE : natural := 16
           );
    port( 
              clk_100_MHz : in  std_logic;
              x_Data      : in  std_logic_vector(DSIZE - 1 downto 0);
              d_Data      : in  std_logic_vector(DSIZE - 1 downto 0);
              e_Data      : out std_logic_vector(DSIZE - 1 downto 0);
              y_Data      : out std_logic_vector(DSIZE - 1 downto 0);
              new_sample  : in  std_logic  
        ); 
end LMS;

architecture Behavioral of LMS is

constant s_rounding : std_logic_vector(2*DSIZE - 1 downto 0) := x"000000004000"; 

type ASOP is array (0 to NBTAP - 1) of signed(2 * DSIZE - 1 downto 0);  
signal a_sop_Q2_38 : ASOP := ( others => x"000000000000" );

type ACOEF is array (0 to NBTAP - 1) of signed(COESIZE - 1 downto 0);
signal a_w_Q1_15 : ACOEF := ( others => x"0000" );

type ACOEF48 is array (0 to NBTAP - 1) of signed(47 downto 0);
signal a_w_Q2_38 : ACOEF48 := ( others => x"000000000000" );

type TAP_LINE is array (0 to NBTAP - 1) of signed(DSIZE-1 downto 0);
signal a_x_Q1_23 : TAP_LINE := ( others => x"000000" );

type fsm_type is(fsm_idle, fsm_getData, fsm_CalcY, fsm_CalcE, fsm_CalcCoe_1, fsm_CalcCoe_2);
signal fsm : fsm_type := fsm_getData;

signal BETA_Q1_15 : std_logic_vector(15 downto 0)            := x"0147"; --0.01*2^15 = 327.68 --> 327
signal s_x_Q1_23  : std_logic_vector(DSIZE - 1 downto 0)     := x"000000";
signal s_d_Q1_23  : std_logic_vector(DSIZE - 1 downto 0)     := x"000000";
signal s_e_Q1_23  : std_logic_vector(DSIZE - 1 downto 0)     := x"000000";
signal s_e_Q1_23_Checked  : std_logic_vector(DSIZE - 1 downto 0)     := x"000000";
signal s_y_Q1_23  : std_logic_vector(DSIZE - 1 downto 0)     := x"000000";
signal s_y_Q2_38  : std_logic_vector(2*DSIZE - 1 downto 0)   := x"000000000000";

signal s_Prod_Q2_38 : std_logic_vector(39 downto 0) := x"0000000000";

begin

LMS : process 
begin
	wait until rising_edge(clk_100_MHz);

        if fsm = fsm_getData then
            if new_sample = '1' then -- --> if rising_edge of new_sample = '1'
                s_x_Q1_23 <= x_Data;
                s_d_Q1_23 <= d_Data;
 
                for i in NBTAP-1 downto 1 loop
                    a_x_Q1_23(i) <= a_x_Q1_23(i-1);
                end loop;  
                    
                fsm <= fsm_CalcY;	
            end if;	
            
        elsif fsm = fsm_CalcY then
             a_x_Q1_23(0)   <= signed(s_x_Q1_23);
            --****************************************************************************
             a_sop_Q2_38(31)   <= signed(s_x_Q1_23) * signed(a_w_Q1_15(31)) + signed(s_rounding);
             
             for I in 1 to NBTAP-1 loop 
                a_sop_Q2_38(NBTAP - I - 1)    <= signed(s_x_Q1_23) * signed(a_w_Q1_15(NBTAP - 1 - I)) + a_sop_Q2_38(NBTAP - I);
             end loop;
            --****************************************************************************

            fsm <= fsm_CalcE;
            
        elsif fsm = fsm_CalcE then  
            s_e_Q1_23   <= std_logic_vector(signed(s_d_Q1_23) - signed(s_y_Q1_23));
            fsm   <= fsm_CalcCoe_1;
            
        elsif fsm = fsm_CalcCoe_1 then
            s_Prod_Q2_38 <= std_logic_vector(signed(BETA_Q1_15) * signed(s_e_Q1_23_Checked));
            fsm <= fsm_CalcCoe_2;
            
        elsif fsm = fsm_CalcCoe_2 then   
            for I in 0 to NBTAP - 1 loop
                a_w_Q2_38(i) <= a_w_Q2_38(i) + signed(s_Prod_Q2_38(38 downto 23)) * a_x_Q1_23(i);--+ signed(s_Prod_Q1_15) * a_x_Q1_23(i);
            end loop;
            fsm <= fsm_getData;
	   end if;	
end process;


s_e_Q1_23_Checked  <=  x"7FFFFF"    when (to_signed(to_integer(signed(s_d_Q1_23)),32) - to_signed(to_integer(signed(s_y_Q1_23)),32)) > x"007FFFFF" else
                       x"800000"    when (to_signed(to_integer(signed(s_d_Q1_23)),32) - to_signed(to_integer(signed(s_y_Q1_23)),32)) < x"FF800000" else 
                       s_e_Q1_23;

s_y_Q1_23 <= std_logic_vector(a_sop_Q2_38(0)(38 downto 15)) when ((a_sop_Q2_38(0)(47 downto 38) =  "0000000000") or  (a_sop_Q2_38(0)(47 downto 38)  = "1111111111" )) else
                                     x"7FFFFF"              when ((a_sop_Q2_38(0)(47)           = '0'          ) and (a_sop_Q2_38(0)(46 downto 38) /= "000000000"  )) else
                                     x"800000";                --((a_sop_Q2_38(0)(47)           = '1'          ) and (a_sop_Q2_38(0)(46 downto 38) /= "111111111"  )) else

gen_l : for I in 0 to NBTAP - 1 generate
a_w_Q1_15(I) <=   a_w_Q2_38(I)(38 downto 23) when ((a_w_Q2_38(I)(47 downto 38) =  "0000000000") or  (a_w_Q2_38(I)(47 downto 38)  = "1111111111" )) else
                           x"7FFF"           when ((a_w_Q2_38(I)(47)           = '0'          ) and (a_w_Q2_38(I)(46 downto 38) /= "000000000"  )) else
                           x"8000";             --((a_w_Q2_38(I)(47)           = '1'          ) and (a_w_Q2_38(I)(46 downto 38) /= "111111111"  )) else
end generate gen_l;

e_Data <= s_e_Q1_23;
y_Data <= s_y_Q1_23;


end Behavioral;
                           
